#ifndef __MATRIX_HH__
#define __MATRIX_HH__

/** @defgroup matrix matrix and row class templates
 *  @brief sparse / dense template matrix with "automaton like" attributes ("final" integer attribute per row)
 */
// @{

//STL
#include <functional>
#include <algorithm>
#include <vector>
#include <map>
//STD
#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
//STR1
#ifdef USE_TYPE_TRAITS
  #include <type_traits>
#else
  #ifdef USE_TR1_TYPE_TRAITS
    #include <tr1/type_traits>
  #else
    #error  <type_traits> or <tr1/type_traits> not selected in "matrix.hh" : try g++ ... -DUSE_TR1_TYPE_TRAITS or -DUSE_TYPE_TRAITS, and use g++ version >= 4.3 with "-std=c++0x" or "-std=gnu++0x"
  #endif
#endif


#ifdef HAS_STD_TYPE_TRAITS
   using namespace std;
#else
   #ifdef HAS_STD_TR1_TYPE_TRAITS
      using namespace std::tr1;
   #else
      #error  std or std::tr1 not selected in "matrix.hh" : try g++ ... -DHAS_STD_TYPE_TRAITS or -DHAS_STD_TR1_TYPE_TRAITS, and use g++ version >= 4.3 with "-std=c++0x" or "-std=gnu++0x"
   #endif
#endif


//STR
#include "macro.h"
#include "cost.hh"

/**@name  Enable/Disable if
 * @brief only for definition and selection of good functions
 */
//@{
/** @brief struct template is enabled for bool = true
 */
template <bool B, typename T = void> struct enable_if_cx { /** type enabled */ typedef T type; };
/** @brief struct template is disabled for bool = false
 */
template <typename T> struct enable_if_cx <false, T> {};

/** @brief struct template is enabled for bool = false
 */
template <bool B, typename T = void>  struct disable_if_cx { /** type enabled */  typedef T type; };
/** @brief struct template is disabled for bool = true
 */
template <typename T> struct disable_if_cx <true, T> {};
//@}



/** Zero and One are defined as "copies" for "arithmetic" types (but not for "Cost" types that are <min,plus> defined ... ),
 *  Transition gives the correct weight for a transition for "arithmetic types" (but for "Cost" type, gives the cost of the letter and not the letter)
 */
//@{
#ifdef HAS_STD_TYPE_TRAITS

/// arithmetic templates (for double)
/** @brief Zero constant in the (+,x) semi-group
 *  @return 0
 */
template<typename T> typename enable_if_cx < std::is_arithmetic<T>::value , T >::type Zero()                   {return T(0);}
/** @brief One constant in the (+,x) semi-group
 *  @return 1
 */
template<typename T> typename enable_if_cx < std::is_arithmetic<T>::value , T >::type One()                    {return T(1);}
/** @brief Probability of a given transition
 *  @param * is the letter of the transition
 *  @param v is the probability of the transition
 *  @return the probability of the transition
 */
template<typename T> typename enable_if_cx < std::is_arithmetic<T>::value , T >::type Transition(int,double v) {return T(v);}

/// cost template (cost is not defined "arithmetic")
/** @addtogroup cost
 */
// @{
/** @brief Zero constant in the (min,+) semi-group
 *  @return +infinity
 */
template<typename T> typename disable_if_cx < std::is_arithmetic<T>::value, T>::type Zero()                   {return T(0x7fffffff);}
/** @brief One constant in the (min,+) semi-group
 *  @return 0
 */
template<typename T> typename disable_if_cx < std::is_arithmetic<T>::value, T>::type One()                    {return T(0x00000000);}
/** @brief Cost of the letter for a given transition
 *  @param a is the letter of the transition
 *  @param * is the probability of the transition
 *  @return the cost of the transition letter
 */
template<typename T> typename disable_if_cx < std::is_arithmetic<T>::value, T>::type Transition(int a,double) {return T(gv_lossless_costs_vector[a]);}
// @}

#else

/// arithmetic templates (for double)
/** @brief Zero constant in the (+,x) semi-group
 *  @return 0
 */
template<typename T> typename enable_if_cx < std::tr1::is_arithmetic<T>::value , T >::type Zero()                   {return T(0);}
/** @brief One constant in the (+,x) semi-group
 *  @return 1
 */
template<typename T> typename enable_if_cx < std::tr1::is_arithmetic<T>::value , T >::type One()                    {return T(1);}
/** @brief Probability of a given transition
 *  @param * is the letter of the transition
 *  @param v is the probability of the transition
 *  @return the probability of the transition
 */
template<typename T> typename enable_if_cx < std::tr1::is_arithmetic<T>::value , T >::type Transition(int,double v) {return T(v);}

/// cost template (cost is not defined "arithmetic")
/** @addtogroup cost
 */
// @{
/** @brief Zero constant in the (min,+) semi-group
 *  @return +infinity
 */
template<typename T> typename disable_if_cx < std::tr1::is_arithmetic<T>::value, T>::type Zero()                   {return T(0x7fffffff);}
/** @brief One constant in the (min,+) semi-group
 *  @return 0
 */
template<typename T> typename disable_if_cx < std::tr1::is_arithmetic<T>::value, T>::type One()                    {return T(0x00000000);}
/** @brief Cost of the letter for a given transition
 *  @param a is the letter of the transition
 *  @param * is the probability of the transition
 *  @return the cost of the transition letter
 */
template<typename T> typename disable_if_cx < std::tr1::is_arithmetic<T>::value, T>::type Transition(int a,double) {return T(gv_lossless_costs_vector[a]);}
// @}

#endif

// @}

/**
 * @class row
 * @brief sparse / non sparce vector, represented by an ordered vector of cells, each cell containing either a couple @f$ ( integer \times T ) @f$ or a single element @f$ T @f$
 * T may represents here
 * - a probability
 * - a cost
 *  each row has a also a flag to together with a "final" state for this row.
 */

template<typename T> class row {

public:

  /** @brief Constructor for an empty row (empty cell lists)
   *  @param final is the row state (final or not)
   *  @param sparse is the implementation choosen for the row being created
   */
  row(const int final = 0, const bool sparse = true): _final(final), _sparse(sparse), _cells_sparse(), _cells_dense() {};

  /** @brief Copy constructor
   *  @param r is the row to be cloned (vector reallocation)
   */
  row(const row<T> & r): _final(r._final), _sparse(r._sparse), _cells_sparse(r._cells_sparse), _cells_dense(r._cells_dense) {};

  /** @brief Erase a row (clear cell lists first)
   */
  ~row() { _cells_sparse.clear(); _cells_dense.clear(); _final = 0; _sparse = true;};

  /** @brief Clear a row (clear the vector _cells and set _final to 0)
   */
  const void clear() { _cells_sparse.clear(); _cells_dense.clear(); _final = 0; _sparse = true;};


  /** @brief add a new cell (j) if does not exists, otherwise "add" (according to T) the "v" value to the current cell (j)
   *  @param j is the column number
   *  @param v is the element to be inserted (or added if it already exists)
   */
  void insert(const int j, const T v);

  /** @brief final state
   *  @return final state of the current row
   */
  const int final() const { return _final; }

  /** @brief set final state
   *  @param final is the value that will be set to the current row
   */
  const void setfinal(const int final) { _final = final; }

  /** @brief get the implementation of the current row
   *  @return true if the row is sparse (dense otherwise)
   */
  const bool sparse() const { return _sparse; }

  /** @brief change the implementation of the current row (sparse to dense and dense to sparse)
   *  @param sparse is the implementation choosen for the row being modified
   */
  const void setsparse(const bool sparse) {
    if (!_sparse && sparse) {
      for (unsigned i = 0; i < _cells_dense.size(); i++) {
        if (_cells_dense[i] != Zero<T>()) {
          pair<int,T> p = make_pair(i,_cells_dense[i]);
          _cells_sparse.push_back(p);
        }
      }
      _cells_dense.clear();
      _sparse = true;
    } else if (_sparse && !sparse) {
      for (typename vector< pair<int,T> >::const_iterator i = _cells_sparse.begin(); i != _cells_sparse.end(); i++) {
        int j = i->first;
        T   v = i->second;
        if (_cells_dense.size() <= (unsigned)j)
          _cells_dense.resize(j+1, Zero<T>());
        _cells_dense[j] = _cells_dense[j] + v;
      }
      _cells_sparse.clear();
      _sparse = false;
    }
  }

  /** @brief return the size as the number of cells
   *  @return the number of cells
   */
  const int size() const { if (_sparse) return _cells_sparse.size(); else return _cells_dense.size(); }

  /** @brief return the size as the number of cells
   *  @return the number of cells
   */
  const int max_index() const {
    if (_sparse) {
      int max_column = 0;
      for (typename vector< pair<int,T> >::const_iterator i = _cells_sparse.begin(); i != _cells_sparse.end(); i++) {
        max_column = MAX(max_column,i->first);
      }
      return max_column;
    } else {
      return _cells_dense.size() - 1;
    }
  }

  /** @brief return the density as the number of cells over the max index (virtual size)
   *  @return the density
   */
  const double density() const {
    if (_sparse) {
      return (double) _cells_sparse.size() / (max_index()+1);
    } else {
      int n = 0;
      for (unsigned i = 0; i < _cells_dense.size(); i++)
        if (_cells_dense[i] != Zero<T>())  n++;
      return (double) n / _cells_dense.size(); //FIXME if (_cell_dense.size() == 0)
    }
  }


  /** @brief return the reference to the vector of sparse cells
   *  @return a reference to the vector
   */
  const vector< pair<int,T> > & cells_sparse() const { return _cells_sparse; }

  /** @brief return the reference to the vector of dense cells
   *  @return a reference to the vector
   */
  const vector< T > & cells_dense() const { return _cells_dense; }

  //@{
  /// print row information
  template<typename U> friend ostream& operator<< (ostream& os, const row<U>& m);
  /// load row information
  template<typename U> friend istream& operator>> (istream& is, row<U>& m);
  //@}

 protected :
  /// final state
  int _final;
  /// is the matrix sparce or not ?
  bool _sparse;
  /// cells list in sparse mode
  vector< pair<int,T> > _cells_sparse;
  /// cells vector in dense mode
  vector< T > _cells_dense;

  /// matrix is a friend class to ease access
  template<typename U> friend class matrix;
  /// matrices_slicer is a friend class to ease access
  template<typename U> friend class matrices_slicer;

};




/**
 * @class matrix
 *
 * @brief sparse matrix, represented by a (dense) vector of (sparse / dense) rows.
 *        This class has strong link with automaton as it usualy represents such objects (final/non-final states) after
 *        being processed with a HMM model.
 * @see row
 * @see automaton
 */

template<typename T> class matrix {

public:


  /** @brief Constructor for an empty matrix
   */

  matrix() : _rows() {};

  /** @brief Copy constructor
   *  @param m is the matrix to be cloned (vector reallocation)
   */

  matrix(const matrix<T> & m) {
    _rows = vector<row<T> >(m._rows.size());
    for (int i = 0; i < (int)m._rows.size(); i++)
      _rows[i] = row<T>(m._rows[i]);
  };


  /** @brief Erase a matrix (clear _cells_sparse and _cells_dense for each row first)
   */
  ~matrix() {
    for (int i = 0; i < (int)_rows.size(); i++)
      _rows[i].clear();
    _rows.clear();
  };

  /** @brief Clear a matrix (clear _cells_sparse and _cells_dense for each row first)
   */
  void clear() {
    for (int i = 0; i < (int)_rows.size(); i++)
      _rows[i].clear();
    _rows.clear();
  };

  /** @brief Resize the number of line of the matrix (clear the rows)
   *  @param s is the new number of rows.
   */
  void resize(int s) {
    _rows.resize(s);
  };

  /** @brief add a new cell (i,j) if does not exists, otherwise "add" (according to T) the "v" value to the current cell (i,j)
   *  @param i is the row number
   *  @param j is the column number
   *  @param v is the element to be inserted (or added if it already exists)
   */
  inline void insert(const int i, const int j, const T v) { _rows[i].insert(j,v);}

  /** @brief add a new row at the end of the matrix
   *  @param final is the row state (final or not)
   *  @param sparse is the implementation choosen for the row being added
   *  @return the "new" index of the row being added
   */
  inline const int addNewRow(const int final = 0, const bool sparse = true) { _rows.push_back(row<T>(final,sparse)); return _rows.size() - 1;}

  /** @brief add a new cell (i,j) if does not exists, otherwise "add" (according to T) the "v" value to the current cell (i,j)
   *  @param i is the row number
   *  @param j is the column number
   *  @param v is the element to be inserted (or added if it already exists)
   */
  inline void addNewCell(const int i, const int j, const T v) { _rows[i].insert(j,v); }

  /** @brief return the size as the number of rows
   *  @return the number of rows
   */
  inline const int size() const { return _rows.size(); }

  /** @brief return the fullsize as the full number of cells
   *  @return the numrber of cells
   */
  inline const int fullsize() const {
    int fsize = 0;
    for (int i = 0; i < (int)_rows.size(); i++)
      fsize += _rows[i].size();
    return fsize;
  }

  /** @brief selfLoop on row i
   *  @param i the row and also the column number (since this is a selfloop)
   *  @param p is the value to be inserted (by default it is a One\<T\>(), i.e. the neutral element of the *\<T\>)
   */
  inline void selfLoop(const int i, const T p = One<T>()) { insert(i,i,p); }


  /** @brief Compute the transposed matrix
   *  @return the transposed matrix
   *  @todo{FIXME : final states/rows are not set}
   */
  matrix<T> * Transpose() const;

  /** @brief Compute the composition of two matrices
   *  @param other is the matrix that will be composed with the current matrix this
   *  @return the composed matrix of the current matrix with the other matrix
   */
  matrix<T> * Compose(const matrix<T> &other) const;

  /** @brief Compute the probability/(min cost) to be at a final/(non final) row during the "nbSteps"th step
   *  @param nbSteps is the number of iterations done on the matrix
   *  @param final is set when the computation is done on final states (otherwise, it is done on non final states)
   *  @return the probability/(min cost) to be at a final/(non final) row during the "nbSteps"-th step
   *  @warning only work with "non window" matrices (must be "square" and "self-injecting") : otherwise final states are not set correctly
   *  @see Pr_one_step_from_one
   */
  const T Pr(const int nbSteps = gv_alignment_length, const bool final = true) const;

  /** @brief Compute a one step single walk from the initial to the final states marked by the matrix "m_final"
   *  @param final is set when the computation is done on final states (otherwise, it is done on non final states)
   *  @param m_final is the matrix used at the end to mark final states
   *    (matrices are otherwise represented as one final/row + a set of links, but nothing is known on the reaching states, so "m_final" is usefull here ...)
   *  @return the probability/(min cost) to be at a final/(non final) row during the "nbSteps"-th step
   *  @warning only work with "full-window" computed matrices (not "self-injecting")
   *  @see Pr
   */
  const T Pr_one_step_from_one(const matrix<T> &m_final, const bool final = true) const;


  /**@name  IO methods
   * @brief IO streams to print/store/load matrix
   */

  //@{
  /// print matrix information
  template<typename U> friend ostream& operator<< (ostream& os, const matrix<U>& m);
  /// load matrix information
  template<typename U> friend istream& operator>> (istream& is, matrix<U>& m);
  //@}

protected :
  /// vector of rows
  vector< row<T> > _rows;

  /// matrices_slicer is a friend class to ease access
  template<typename U> friend class matrices_slicer;
};


/// compare pairs of elements in a row only on their first index number (and not the second T element)
template<typename T> inline bool pairless(const pair<int, T> l, const pair<int, T> r) {
  return l.first < r.first;
}


/// output method for a row
template<typename T> inline ostream& operator<<(ostream& os, const row<T>& r) {
  os << "\t" << (r.size()) << "\t" << (r.final()) << endl;
  // for each row, display each cell
  if (r.sparse()) {
    for (typename vector< pair<int,T> >::const_iterator i = r._cells_sparse.begin(); i != r._cells_sparse.end(); i++) {
      os << "\t\t\t" << (i->first) << "\t" << (i->second) << endl;
    }
  } else {
    for (typename vector< T >::const_iterator i = r._cells_dense.begin(); i != r._cells_dense.end(); i++) {
      if ((*i) != Zero<T>()) {
        os << "\t\t\t" << (i - r._cells_dense.begin()) << "\t" << (*i) << endl;
      }
    }
  }
  return os;
}


/// input method for a row
template<typename T> inline istream& operator>>(istream& is, row<T>& r) {
  // previous data removed if any
  r._cells_dense.clear();
  r._cells_sparse.clear();
  r._sparse = true;

  // a) read row length
  int nbcells = 0;
  is >> nbcells;
#ifdef DEBUGREADING
  cerr << nbcells << endl;
#endif

  if (nbcells <= 0) {
    cerr << "> when reading nbcells" << endl;
    _ERROR("row operator>>","incorrect size "<< nbcells);
  }

  // b) read row final
  int final;
  is >> final;
  r._final = final;

  // c) read row data
  for (int j = 0; j<nbcells; j++){
    int  to = 0;
    T    pr = T();
    is >> to >> pr;
#ifdef DEBUGREADING
    cerr << "\t\t" << to << "\t" << pr << endl;
#endif
    if (to < 0) {
      cerr << "> when reading cell" << endl;
      _ERROR("row operator>>","incorrect column value "<< to);
    }
    r.insert(to,pr);
  }
  return is;
}


/// output method for the current matrix
template<typename T> inline ostream& operator<<(ostream& os, const matrix<T>& m) {
  os << m._rows.size() << endl;
  // display each row
  for (typename vector< row<T> >::const_iterator i = m._rows.begin(); i != m._rows.end(); i++) {
    const row<T> & r = *i;
    os << "\t" << r;
  }
  return os;
}

/// input method for the current matrix
template<typename T> inline istream& operator>>(istream& is, matrix<T>& m) {
  // clear previous matrix
  m._rows.clear();

  // read matrix size
  int nbrows = 0;
  is >> nbrows;
#ifdef DEBUGREADING
  cerr << nbrows << endl;
#endif
  if (nbrows <= 0) {
    cerr << "> when reading matrix size" << endl;
      _ERROR("matrix operator>>","incorrect size "<< nbrows);
  }

  // add rows first
  for (int i = 0;i<nbrows;i++){
    m.addNewRow();
  }

  // get each row content
  int introw;
  for (int from = 0; from<nbrows; from++){
    is >> introw >> (m._rows[from]);

#ifdef DEBUGREADING
    cerr << "\t"  << introw << "\t" << (m._rows[from]);
#endif

    if (introw < 0 || introw >= nbrows) {
      cerr << "> when reading matrix cell line " << introw << " (line nb " << from << ")" << endl;
      _ERROR("matrix operator>>","incorrect row number " <<  introw << " (not in [0.."<<(nbrows-1)<<"])");
    }
  }
  return is;
}




/// insert an element to a row (or "add" it using the T "+" operator)
template<typename T> inline void row<T>::insert(const int j, const T v) {
  if (_sparse) {
    /* sparse implementation */
    pair<int,T> p(j,v);
    typename vector< pair<int,T> >::iterator it = lower_bound(_cells_sparse.begin(),_cells_sparse.end(),p,pairless<T>);
    if (it != _cells_sparse.end())
      if (it->first == j)
        it->second = it->second + v;
      else
        _cells_sparse.insert(it,p);
    else
      _cells_sparse.push_back(p);
  } else {
    /* dense implementation */
    if (_cells_dense.size() <= (unsigned)j)
      _cells_dense.resize(j+1, Zero<T>());
    _cells_dense[j] =  _cells_dense[j] + v;
  }
}



/// transpose a matrix
template<typename T> inline matrix<T> * matrix<T>::Transpose() const {
  matrix * result = new matrix();

  // find the max column on the original matrix
  int max_column = 0;

  for (int i = 0; i < (int)_rows.size(); i++) {
    max_column = MAX(max_column,_rows[i].max_index());
  }

  // build the transpose matrix in dense format first
  for (int i = 0; i <= max_column; i++) {
    result->addNewRow(false/* final does not mean anything here so set to false */, false /* dense */);
    result->_rows[i]._cells_dense.resize(_rows.size(), Zero<T>());
  }

  // fill the transpose matrix in dense format first then put the sparse row when needed ...
  for (unsigned i = 0; i < _rows.size(); i++) {
    if (_rows[i]._sparse) {
      for (typename vector< pair<int,T> >::const_iterator j=_rows[i]._cells_sparse.begin(); j != _rows[i]._cells_sparse.end(); j++) {
        result->_rows[j->first]._cells_dense[i] = j->second;
      }
    } else {
      for (unsigned j = 0 ; j < _rows[i]._cells_dense.size(); j++) {
        result->_rows[j]._cells_dense[i] = _rows[i]._cells_dense[j];
      }
    }
  }

  for (int i = 0; i < (int)result->_rows.size(); i++) {
    if (result->_rows[i].density() < MATRIX_SPARSE_ROW_DENSITY) {
      result->_rows[i].setsparse(true);
    }
  }
  return result;
}





/// matrix composition (looks like a product)
template<typename T> inline matrix<T> * matrix<T>::Compose(const matrix<T> &other) const {
  matrix * result = new matrix();
  matrix * other_transpose = other.Transpose();

  for (int i = 0; i < (int)_rows.size(); i++) {
    result->addNewRow(_rows[i].final(),false);
    result->_rows[i]._cells_dense.resize(other_transpose->_rows.size(), Zero<T>()); // FIXME dense mode first
    for (int j = 0; j < (int)other_transpose->_rows.size(); j++) {
      if (_rows[i]._sparse) {
        if (other_transpose->_rows[j]._sparse) {
          /* 1/4 */
          typename vector< pair<int,T> >::const_iterator k_this =                   _rows[i]._cells_sparse.begin();
          typename vector< pair<int,T> >::const_iterator k_other = other_transpose->_rows[j]._cells_sparse.begin();
          while (k_this != _rows[i]._cells_sparse.end() && k_other != other_transpose->_rows[j]._cells_sparse.end()) {
            if (k_this->first == k_other->first) {
              result->_rows[i]._cells_dense[j] = result->_rows[i]._cells_dense[j] + (k_this->second * k_other->second);
              k_this++;
              k_other++;
            } else if (k_this->first < k_other->first) {
              k_this++;
            } else {
              k_other++;
            }
          }
        } else {
          /* 2/4 */
          for (typename vector< pair<int,T> >::const_iterator k_this = _rows[i]._cells_sparse.begin(); k_this < _rows[i]._cells_sparse.end(); k_this++) {
            if (other_transpose->_rows[j]._cells_dense[k_this->first] != Zero<T>()) {
              result->_rows[i]._cells_dense[j] = result->_rows[i]._cells_dense[j] + (k_this->second * other_transpose->_rows[j]._cells_dense[k_this->first]);
            }
          }
        }
      } else {
        if (other_transpose->_rows[j]._sparse) {
          /* 3/4 */
          for (typename vector< pair<int,T> >::const_iterator k_other = other_transpose->_rows[j]._cells_sparse.begin(); k_other < other_transpose->_rows[j]._cells_sparse.end(); k_other++) {
            if (_rows[i]._cells_dense[k_other->first] != Zero<T>()) {
              result->_rows[i]._cells_dense[j] = result->_rows[i]._cells_dense[j] + (_rows[i]._cells_dense[k_other->first] * k_other->second);
            }
          }
        } else {
          /* 4/4 */
          for (unsigned k=0; k < _rows[i]._cells_dense.size(); k++) {
            result->_rows[i]._cells_dense[j] = result->_rows[i]._cells_dense[j] + (_rows[i]._cells_dense[k] * other_transpose->_rows[j]._cells_dense[k]);
          }
        }
      }
    }
    if (result->_rows[i].density() < MATRIX_SPARSE_ROW_DENSITY) {
      result->_rows[i].setsparse(true);
    }
  }

  delete other_transpose;
  return result;
}






/// matrix "Pr" classical algorithm : compute the "T"-lity to be at a final or non final state after nbSteps on the same matrix
template<typename T> inline const T matrix<T>::Pr(const int nbSteps, const bool final) const {
  int i_min = 1;
  int i_max = 1;

  // set all the other probabilities to 0.0 (plus neutral element) or cost to infinity (MIN neutral element)
  vector<T> v0(size(),Zero<T>());
  vector<T> v1(size(),Zero<T>());

  // set initial state probability to 1.0 (product neutral element) or initial cost to 0 (PLUS neutral element)
  v1[1] = One<T>();

  // compute probability/cost at step i provided probabilities/costs at step i-1
  for (int k = 0 ; k < nbSteps ; k++) {
    int j_min = size()-1;
    int j_max = 0;
    if (k&1) {
      for (int i = i_min; i <= i_max; i++) {
        if (v0[i] != Zero<T>()) { // p is not the "+" neutral element (0 for prob, or infinity for costs)
          if (_rows[i]._sparse) {
            for (typename vector< pair<int,T> >::const_iterator j=_rows[i]._cells_sparse.begin(); j != _rows[i]._cells_sparse.end(); j++) {
              j_min = MIN(j_min,j->first);
              j_max = MAX(j_max,j->first);
              v1[j->first] = v1[j->first] + ((j->second) * (v0[i]));
            }
          } else {
            for (unsigned j = 0 ; j < _rows[i]._cells_dense.size(); j++) {
              if (_rows[i]._cells_dense[j] != Zero<T>()) {
                j_min = MIN(j_min,(int)j);
                j_max = MAX(j_max,(int)j);
                v1[j] = v1[j] + ((_rows[i]._cells_dense[j]) * (v0[i]));
              }
            }
          }
          v0[i] = Zero<T>();
        }
      }
    } else {
      for (int i = i_min; i <= i_max; i++) {
        if (v1[i] != Zero<T>()) { // p is not the "+" neutral element (0 for prob, or infinity for costs)
          if (_rows[i]._sparse) {
            for (typename vector< pair<int,T> >::const_iterator j = _rows[i]._cells_sparse.begin(); j != _rows[i]._cells_sparse.end(); j++) {
              j_min = MIN(j_min,j->first);
              j_max = MAX(j_max,j->first);
              v0[j->first] = v0[j->first] + ((j->second) * (v1[i]));
            }
          } else {
            for (unsigned j = 0 ; j < _rows[i]._cells_dense.size(); j++) {
              if (_rows[i]._cells_dense[j] != Zero<T>()) {
                j_min = MIN(j_min,(int)j);
                j_max = MAX(j_max,(int)j);
                v0[j] = v0[j] + ((_rows[i]._cells_dense[j]) * (v1[i]));
              }
            }
          }
          v1[i] = Zero<T>();
        }
      }
    }
    i_min = j_min;
    i_max = j_max;
  } // for


  T result = Zero<T>();
  // Sum "final" states probabilities or find the Min "non final" cost (NB : (+) ~ min )
  if (final) {
    if (nbSteps & 1) {
      for (int i=i_min;i<=i_max;i++)
        if (_rows[i].final())
          result = (result + (v0[i]));
    } else {
      for (int i=i_min;i<=i_max;i++)
        if (_rows[i].final())
          result = (result + (v1[i]));
    }
  } else {
    if (nbSteps & 1) {
      for (int i=i_min;i<=i_max;i++)
        if (!(_rows[i].final()))
          result = (result + (v0[i]));
    } else {
      for (int i=i_min;i<=i_max;i++)
        if (!(_rows[i].final()))
          result = (result + (v1[i]));
    }
  }

  v0.clear();
  v1.clear();
  return result;
}

/// matrix "Pr_one_step_from_one" algorithm : compute the "T"-lity to be at a final or non final state after one step from 1 in the original matrix to the m_final one
template<typename T> inline const T matrix<T>::Pr_one_step_from_one(const matrix<T> &m_final, const bool final) const {
  T result = Zero<T>();
  // Sum "final" states probabilities or find the Min "non final" cost (NB : (+) ~ min )
  if (final) {
    if (_rows[1]._sparse) {
      for (typename vector< pair<int,T> >::const_iterator j = _rows[1]._cells_sparse.begin(); j != _rows[1]._cells_sparse.end(); j++)
        if (m_final._rows[j->first]._final)
          result = (result + (j->second));
    } else {
      for (unsigned int j = 0 ; j < _rows[1]._cells_dense.size(); j++) {
        if (m_final._rows[j]._final)
          result = (result + (_rows[1]._cells_dense[j]));
      }
    }
  } else {
    if (_rows[1]._sparse) {
      for (typename vector< pair<int,T> >::const_iterator j = _rows[1]._cells_sparse.begin(); j != _rows[1]._cells_sparse.end(); j++)
        if (!(m_final._rows[j->first]._final))
          result = (result + (j->second));
    } else {
      for (unsigned int j = 0 ; j < _rows[1]._cells_dense.size(); j++) {
        if (!(m_final._rows[j]._final))
          result = (result + (_rows[1]._cells_dense[j]));
      }
    }
  }
  return result;
}
// @}







/** @addtogroup matrix
 */
// @{


/** @brief Compute the slicer structure needed to quickly computer overlapped windows sensitivity :
 *       @li _i and _j are the left/right indices of the window
 *       @li _matrices_data keep the matrices (and update them to keep a "binary-trie" structure that fasten the possible products between _i and _j)
 *       @li _left/_right matrices block and index are also there to help in the computation of successive windows :
 *           _left is maintained as a stack of matrices whereas _right is a single matrix.
*
 */

template<typename T> class matrices_slicer {

public :

  /** @brief constructor for a slicer from the set of matrices obtained with a "slice" (or "step") product
   *  @param data is the vector of sliced matrices been used (and erased when finished)
   *  @see  matrices_step_pr_product
   *  @see  matrices_step_cost_product
   */
  matrices_slicer(vector< matrix<T> * > * data): _matrices_data(data), _i(0), _j(0), _middle(0) {
    _left_matrices_block = stack<matrix<T>*>();
    _left_matrices_index = stack<int>();
    _right_matrix_block = NULL;
    _right_matrix_index = 0;
#ifdef MATRIX_SLICER_STATS
    nb_prod_a           = 0;
    nb_prod_b           = 0;
    nb_prod_c           = 0;
    nb_prod_d           = 0;
    nb_middle_change    = 0;
#endif
#ifndef NO_FULL_MATRIX_BLOCK
    _full_matrix_block   = NULL;
    _full_matrix_block_i = -1;
    _full_matrix_block_j = -1;
#ifdef MATRIX_SLICER_STATS
    nb_prod_e = 0;
#endif
#endif

  };


  /** destructor for a slicer
   */
  ~matrices_slicer() {
    // empty stacks
    empty_stacks();
    // delete input data
    for (int i=_i;i<(int)_matrices_data->size();i++) {
      (*_matrices_data)[i]->clear();
      delete (*_matrices_data)[i];
    }
    _matrices_data->clear();
    delete _matrices_data;
    _matrices_data = NULL;

    _middle        = 0;
    _i             = 0;
    _j             = 0;
  }



  /** @brief empty stacks
   */
  inline void empty_stacks() {
    while(!_left_matrices_index.empty()) {
      _left_matrices_index.pop();
      delete _left_matrices_block.top();
      _left_matrices_block.pop();
    }
    _right_matrix_index = 0;
    if (_right_matrix_block) {
      delete _right_matrix_block;
      _right_matrix_block = NULL;
    }
  };



  /** @brief right_size
   *  "right" size of a block starting at "s" in the dataset and that must be < "j"
   *  @param s is the starting position of the block
   *  @param j is the right position that must be not crossed (e.g. put "_j+1")
   *  @return the length of the block starting at "s"
   *  @todo{FIXME: check the function due to s}
   */
  inline int right_size(const int s, const int j) const {
    int u = s;
    int t = 1; // 1,2,4,8
    while ((u % 2 == 0) && (s + (t << 1) <= j )) {
      t  = t << 1;
      u  = u >> 1;
    }
    return t;
  }

  /** @brief left_size
   *  "right" size of a block finishing at "s-1" in the dataset and that must be >= "i"
   *  @param i is the left position that can be crossed but not sub-overlapped
   *  @param s is the ending + 1 position of the block (e.g. put "_j+1")
   *  @return the length of the block ending at "s-1"
   *  @todo{FIXME: check the function due to s}
   */

  inline int left_size(const int i, const int s) const {
    int u = s;
    int t = 1; // 1,2,4,8
    while ((u % 2 == 0) && (i + (t << 1) <= s)) {
      t     = t << 1;
      u     = u >> 1;
    }
    return t;
  }

  /** @brief middle_pos_right
   *  @param i is the left pos
   *  @param j is the right pos
   *  @return the end " + 1" of the maximal block included withing i and < j
   *  @todo{FIXME: check the function due to s}
   */
  inline int middle_pos_right(const int i, const int j) const {
    int s = i;
    int u = 0;
    do {
      int un = right_size(s,j);
      if (un > u) {
        u = un;
        s += u;
      } else {
        return s;
      }
    } while (1);
    return s;
  };

  /** @brief middle_pos_left
   *  @param i is the left pos
   *  @param j is the right pos
   *  @return the beginning of the maximal block included withing i and < j
   *  @todo{FIXME: check the function due to s}
   */
  inline int middle_pos_left(const int i, const int j) const {
    int s = j;
    int u = 0;
    do {
      int un = left_size(i,s);
      if (un > u) {
        u = un;
        s -= u;
      } else {
        return s;
      }
    } while(1);
    return s;
  };

  /** @brief increment _j, add the rightmost matrix _matrices_data[_j] (already done) and update matrices blocks that now include _matrices_data[_j]
   */
  inline void add_right() {
    _j++;
    {
      int u     = _j+1;
      int t_old = 0;
      int t     = 1; // 1,3,7,15 ...
      while ((u % 2 == 0) && (_j - t >= _i)) {
        matrix<T> * tmp_j_t       = (*_matrices_data)[_j - t];
        (*_matrices_data)[_j - t] = tmp_j_t->Compose(*((*_matrices_data)[_j - t_old]));
#ifdef MATRIX_SLICER_STATS
        nb_prod_a++;
#endif
        delete      tmp_j_t;
        t_old = t;
        t    += t + 1;
        u     = u >> 1;
      }
    }
    // erase when middle changes
    int n_middle = middle_pos_left(_i,_j+1);
    if (n_middle != _middle) {
      // FIXME dont empty stack by reusing them
      empty_stacks();
      _middle = n_middle;
#ifdef MATRIX_SLICER_STATS
      nb_middle_change++;
#endif
    }
  }


  /** @brief delete the leftmost  matrix and increment _i
   */
  inline void del_left() {
    if (_i < _j) {
      delete (*_matrices_data)[_i];
      (*_matrices_data)[_i] = NULL;
      _i++;
      // erase when middle changes
      int n_middle = middle_pos_left(_i,_j+1);
      if (n_middle != _middle) {
        empty_stacks();
        _middle = n_middle;
#ifdef MATRIX_SLICER_STATS
        nb_middle_change++;
#endif
      }
    } else {
      _WARNING("[matrix.hh] matrices_slicer.del_left()","_i cannot be > _j : del_left() is ignored.");
    }
  }


  /** @brief  compute the current windows "probability/minimal cost" obtained from
   *          the matrix product from _i to _j
   *  @param  final (default true) select if final or non final states have to be reached.
   *  @return a "probability/minimal cost" depending on the T template
   */

  inline const T current_Pr(const bool final = true) {
    matrix<T> *     left_result  = NULL;
    matrix<T> *     right_result;
    matrix<T> * left_right_result = NULL;

#ifdef MATRIX_SLICER_BUILDPROD
    cerr << (*this) << endl;
    cerr << "* current_product(" << _i << "," << _j << ")  {middle = " << _middle << "}" << endl;
#endif


    // 0) Full matrix block
#ifndef NO_FULL_MATRIX_BLOCK

    if ((_full_matrix_block != NULL) && (_i == _full_matrix_block_i)) {
      // a) use the previous computation when _i has not changed ...
      int lj = right_size(_full_matrix_block_j,_j+1);
      while (_full_matrix_block_j + lj <= _j+1) { // FIXME  : <=  +1?
#ifdef MATRIX_SLICER_BUILDPROD
        cerr << "    MB(compose) -> = " << _full_matrix_block_j << " + " << lj << endl;
#endif
        matrix<T> * old_full_matrix_block = _full_matrix_block;
                       _full_matrix_block = _full_matrix_block->Compose(*((*_matrices_data)[_full_matrix_block_j]));
        delete      old_full_matrix_block;

#ifdef MATRIX_SLICER_STATS
        nb_prod_e++;
#endif

        _full_matrix_block_j += lj;
        lj = right_size(_full_matrix_block_j,_j+1);
      }
      // and jump to te computation
      goto full_matrix_reuse;
    } else {
      // b) delete since full_block will not be usefull when _i has changed ...
      _full_matrix_block_i = -1;
      _full_matrix_block_j = -1;
      if (_full_matrix_block)
        delete _full_matrix_block;
      _full_matrix_block = NULL;
    }
#endif




    // I) Right
    {
      int j_stack;
      matrix<T> * old_right_result = NULL;

      // a) get right border already computed from _middle if it exists ...
      if (_right_matrix_block != NULL) {
        right_result = _right_matrix_block;
        j_stack      = _right_matrix_index;

#ifdef MATRIX_SLICER_BUILDPROD
        cerr << " R(from index) -> = " << j_stack << endl;
#endif

      } else {
        right_result     = new matrix<T>(*((*_matrices_data)[_middle]));
        j_stack          = _middle + right_size(_middle,_j+1);

        old_right_result = right_result;

#ifdef MATRIX_SLICER_BUILDPROD
        cerr << "        _middle -> = " << _middle << ", _middle_length -> = " << (right_size(_middle,_j+1)) << endl;
        cerr << " R(from middle) -> = " << j_stack << endl;
#endif
      }

      // b) compute the right part and push it to the right stack ...
      int lj = right_size(j_stack,_j+1);
      while (j_stack + lj <= _j+1) { // FIXME  : <=  ?

#ifdef MATRIX_SLICER_BUILDPROD
        cerr << " R(compose) -> = " << j_stack << " + " << lj << endl;
#endif

        right_result = right_result->Compose(*((*_matrices_data)[j_stack]));
#ifdef MATRIX_SLICER_STATS
        nb_prod_b++;
#endif

        if (old_right_result)
          delete old_right_result;
        old_right_result = right_result;

        j_stack += lj;
        lj = right_size(j_stack,_j+1);
      }

      // c) update right_matrix_block when needed
      if (_right_matrix_block == right_result) {
        // NO CHANGE !!
#ifdef MATRIX_SLICER_BUILDPROD
        cerr << " R(finalize) -> = NO CHANGE " << j_stack << "=" << _right_matrix_index << endl;
#endif
      } else {
        // SOME CHANGE !!
#ifdef MATRIX_SLICER_BUILDPROD
        cerr << " R(finalize) -> = CHANGE from " << _right_matrix_index << " to " << j_stack << endl;
#endif
        if (_right_matrix_block != NULL)
          delete _right_matrix_block;
        _right_matrix_block = right_result;
        _right_matrix_index = j_stack;
      }
    }

    // II) Left

    // a) get left border already computed if exists ...
    {
      int i_stack;
      while ((!_left_matrices_index.empty()) && ((i_stack = _left_matrices_index.top()) < _i)) {
        _left_matrices_index.pop();
        matrix<T> * t = _left_matrices_block.top();
        _left_matrices_block.pop();
        delete t;
#ifdef MATRIX_SLICER_BUILDPROD
        cerr << " L(stack) -> popping " << i_stack<< endl;
#endif
      }

      if (!_left_matrices_index.empty()) {
        left_result = _left_matrices_block.top();
        i_stack     = _left_matrices_index.top();
#ifdef MATRIX_SLICER_BUILDPROD
        cerr << " L(stack) -> Search success at " << i_stack << endl;
#endif
      } else {
        i_stack = _middle;
#ifdef MATRIX_SLICER_BUILDPROD
        cerr << " L(stack) -> Search failure (empty stack after search)" << endl;
#endif
      }


      // b) compute (and store) the rest of the left elements ...
      int li = left_size(_i,i_stack);
      while (i_stack - li >= _i) {

#ifdef MATRIX_SLICER_BUILDPROD
        cerr << " L(compose) -> = " << i_stack << "-" << li << " = " << (i_stack - li) << endl;
#endif

        i_stack -= li;

        if (left_result) {
          left_result = (*_matrices_data)[i_stack]->Compose(*left_result);
#ifdef MATRIX_SLICER_STATS
          nb_prod_c++;
#endif

        } else {
          left_result   = new matrix<T>(*((*_matrices_data)[i_stack]));
        }
#ifdef MATRIX_SLICER_BUILDPROD
        cerr << " L(push) -> = " << i_stack << endl;
#endif

        _left_matrices_index.push(i_stack);
        _left_matrices_block.push(left_result);

        li = left_size(_i,i_stack);
      }
    }


    // III) Combine Left and Right
    if (left_result) {

#ifdef MATRIX_SLICER_BUILDPROD
      cerr << " M(join)" << endl;
#endif
      left_right_result = left_result->Compose(*right_result);
#ifdef MATRIX_SLICER_STATS
      nb_prod_d++;
#endif

    } else {
      left_right_result = right_result;
#ifdef MATRIX_SLICER_BUILDPROD
      cerr << " M(free join)" << endl;
#endif
    }


    // IV) Full_Matrix "Reuse"
#ifndef NO_FULL_MATRIX_BLOCK
    if (_full_matrix_block != NULL) {
    full_matrix_reuse:
      left_right_result = _full_matrix_block;
    } else {
      if (left_result)
        _full_matrix_block = left_right_result;
      else
        _full_matrix_block = new matrix<T> (*left_right_result); // in fact "left_right_result"="right_result" here, so must be copied to avoid "double free"
      _full_matrix_block_i = _i;
      _full_matrix_block_j = _j+1;
    }
#endif



    // V) Compute sensitivity/lossless property
    T result = Zero<T>();
    matrix<T> * _m_final = (*_matrices_data)[_j+1];

    // Sum "final" states probabilities or find the Min "non final" cost (NB : (+) ~ min )
    if (final) {
      if (left_right_result->_rows[1]._sparse) {
        for (typename vector< pair<int,T> >::const_iterator j = left_right_result->_rows[1]._cells_sparse.begin(); j != left_right_result->_rows[1]._cells_sparse.end(); j++)
          if (_m_final->_rows[j->first]._final)
            result = (result + (j->second));
      } else {
        for (unsigned j = 0 ; j <  left_right_result->_rows[1]._cells_dense.size(); j++) {
          if (_m_final->_rows[j]._final)
            result = (result + (left_right_result->_rows[1]._cells_dense[j]));
        }
      }
    } else {
      if (left_right_result->_rows[1]._sparse) {
        for (typename vector< pair<int,T> >::const_iterator j = left_right_result->_rows[1]._cells_sparse.begin(); j != left_right_result->_rows[1]._cells_sparse.end(); j++)
          if (!(_m_final->_rows[j->first]._final))
            result = (result + (j->second));
      } else {
        for (unsigned j = 0 ; j <  left_right_result->_rows[1]._cells_dense.size(); j++) {
          if (!(_m_final->_rows[j]._final))
            result = (result + (left_right_result->_rows[1]._cells_dense[j]));
        }
      }
    }
// delete if not used
#ifdef NO_FULL_MATRIX_BLOCK
   if (left_result)
     delete left_right_result;
#endif
   return result;
}



  /**@name  IO methods
   * @brief IO streams to print/store/load matrix
   */

  // @{
  /// print matrices_slicer information
  template<typename U> friend ostream& operator<< (ostream& os, const matrices_slicer<U>& ms);
  // @}

#ifdef MATRIX_SLICER_STATS
  /** @name statistics of the number of products done within the matrix_slicer
   */
  // @{
  unsigned              nb_prod_a,nb_prod_b,nb_prod_c,nb_prod_d,nb_middle_change;
#ifndef NO_FULL_MATRIX_BLOCK
  unsigned              nb_prod_e;
#endif
  // @}
#endif

  /** @brief data to be sliced (will be modified so must not be !!!)
   *
   */
  vector< matrix<T> * >  * _matrices_data;
  /// left window index (del only)
  int _i;
  /// right window index (add only)
  int _j;
  /// separator left/right = defined by (2^k) * r such that k is the largest integer (and r is a odd integer obviously) : i <= (2^k) * r <= j
  int _middle;
  /// left matrices (when preprocessed before...)
  // @{
  stack< matrix<T> * >  _left_matrices_block;
  stack< int >          _left_matrices_index;
  // @}
  /// right matrix (when preprocessed before...)
  // @{
  matrix<T> *           _right_matrix_block;
  int                   _right_matrix_index;
  // }@
#ifndef NO_FULL_MATRIX_BLOCK
  /// full block (when preprocessed before...)
  // @{
  matrix<T> *           _full_matrix_block;
  int                   _full_matrix_block_i;
  int                   _full_matrix_block_j;
  // @}
#endif
};




/// output method for the current matrix
template<typename T> inline ostream& operator<<(ostream& os, const matrices_slicer<T>& ms) {
  os << "i = " << (ms._i) << "\t" << "j = " << (ms._j) << "\t" << "m = " << (ms._middle) << "{" << (ms.middle_pos_left(ms._i,ms._j+1)) << "," << (ms.middle_pos_right(ms._i,ms._j+1)) << "}" << endl;

#ifdef MATRIX_SLICER_STATS
  os << (ms.nb_prod_a) << "\t" << (ms.nb_prod_b) << "\t" << (ms.nb_prod_c) << "\t" << (ms.nb_prod_d) << "\t=\t" <<
    (
     (ms.nb_prod_a)
     +
     (ms.nb_prod_b)
     +
     (ms.nb_prod_c)
     +
     (ms.nb_prod_d)
#ifndef NO_FULL_MATRIX_BLOCK
     +
     (ms.nb_prod_e)
#endif
     ) << "\t" << (ms.nb_middle_change) << "\t" << endl;
#endif

  if (ms._right_matrix_block)
    os << "  _right : " << (ms._right_matrix_index) << endl;
  else
    os << "  _right : unset" << endl;

  if (ms._left_matrices_index.size() > 0)
    os << "  _left  : "  << (ms._left_matrices_index.top()) << endl;
  else
    os << "  _left  : empty" << endl;

  os << "  _matrices_data[] (size) : ";
  for (int i = ms._i; i <= ms._j; i++) {
    os << " m[" << i << "] : " << (((*(ms._matrices_data))[i])->size()) << "{" << (((*(ms._matrices_data))[i])->fullsize()) << "} ; ";
  }
  os << endl;
  // FIXME to be continued
  return os;
}




// @}

#endif
